! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
program testbkgoptrateaux01
  ! Authors:  Masaaki Sakano - SSC/LUX
  ! NOTE: 
  ! TODO: (1) test the cases with FRACEXP=0 for some bins in test_findOptimumY().

  use errorhandling
  use types
  use dal		! 

  use bkgoptrate_aux

  ! Followings are in ssclib
  !use constants		! for Pi
  use ssc_misc		! for isNaN
  use test_utils	! for isNearlyEqual()

  implicit none

!   character(36), parameter :: &
!        OUTFILNAME  = "outfiles/testinput_for_getstrs.fits" &
!      , OUTFILNAME2 = "outfiles/testinput_for_retkwd2.fits" &
!      , OUTFILNAME3 = "outfiles/testinput_for_retkwd3.fits"
!   character(80), parameter :: &
!        INCATFILENAME = "testdata/testcat.fits"
!   character(10), parameter :: TMPTBLNAME = "test_table"
! 
!   integer, parameter :: MAINDEFARRAYSIZE = 3
!   real(single), dimension(MAINDEFARRAYSIZE) :: srcRadEffExample
!   real(single), dimension(MAINDEFARRAYSIZE, MAINDEFARRAYSIZE) :: &
!        sepExample, oveFracExample

  !-------------------------------------------------------------------
  ! Start:

  call test_getSN2ofTs()
  call test_findOptimumY()
  call test_calcThresholdFromAny()
  call test_calcThresholdFromRate()

contains

  !---------------------------------------------------------------------
  subroutine test_getSN2ofTs()
  !---------------------------------------------------------------------
  ! Tests of getSN2ofTs()
  ! NOTE: 

    implicit none

    character(64)  :: errHead
    integer, parameter :: defArraySize = 5, DefPrecision = 6

    real(single), dimension(defArraySize) :: tSeries, aryExp
    logical,      dimension(defArraySize) :: mask
    real(single) :: f

    !-------------------------------------------------------------------
    ! Start:

    !----------------------------------------------------------
    ! getSN2ofTs()
    !----------------------------------------------------------
    errHead = "getSN2ofTs001a"
    print *,"=== "//trim(errHead)//": Simple (no mask)"

    ! Prepare a set of example test data
    tSeries(:) = (/ 0.0, 1.0, 1.0, 2.0,  3.0 /)

    if (.not. isNearlyEqual(getSN2ofTs(tSeries), 25/7.0, DefPrecision)) then
      print *, 'getSN2ofTs() is ', getSN2ofTs(tSeries), ' which should be ', 25/7.0
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "getSN2ofTs002a"
    print *,"=== "//trim(errHead)//": Simple (with mask)"

    ! Prepare a set of example test data
    mask(:) = .true.
    mask(2) = .false.
    ! tSeries(:) = (/ 0.0, 1.0, 1.0, 2.0,  3.0 /)
    !    mask(:) =     T    F    T    T     T

    if (.not. isNearlyEqual(getSN2ofTs(tSeries, mask), 16/6.0, DefPrecision)) then
      print *, 'getSN2ofTs() is ', getSN2ofTs(tSeries, mask), ' which should be ', 16/6.0
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "getSN2ofTs002b"
    print *,"=== "//trim(errHead)//": Simple (with null mask)"

    ! Prepare a set of example test data
    mask(:) = .false.
    ! tSeries(:) = (/ 0.0, 1.0, 1.0, 2.0,  3.0 /)
    !    mask(:) =     F    F    F    F     F

    f = getSN2ofTs(tSeries, mask)
    if (f .ne. 0.0) then
    ! if (.not. isNaN(f)) then	! isNaN() defined in ssc_misc
      ! OK
      print *, 'getSN2ofTs() is ', f, ' which should be 0.0.'
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "getSN2ofTs003a"
    print *,"=== "//trim(errHead)//": Simple (with negative)"

    ! Prepare a set of example test data
    tSeries(:) = (/ 0.0, 3.0, 2.0, 1.0, -1.0 /)
    mask(:) = .true.
    mask(2) = .false.
    !  mask(:) =     T    F    T    T     T

    if (.not. isNearlyEqual(getSN2ofTs(tSeries, mask), 16/2.0, DefPrecision)) then	! So far, negative values are simply added.  This behaviour may change in the future...
      print *, 'getSN2ofTs() is ', getSN2ofTs(tSeries, mask), ' which should be ', 16/2.0
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "getSN2ofTs004a"
    print *,"=== "//trim(errHead)//": Simple (without mask but with aryExp)"

    ! Prepare a set of example test data
    tSeries(:) = (/ 0.0, 1.0, 1.0, 2.0,  3.0 /)
    aryExp(:) = 1.0
    aryExp(2) = 0.0
    !aryExp(:) = (/ 1.0, 0.0, 1.0, 1.0,  1.0 /)
    !  mask(:) =>    T    F    T    T     T	! NOT given

    if (.not. isNearlyEqual(getSN2ofTs(tSeries, aryExp=aryExp), 16/6.0, DefPrecision)) then
      f = getSN2ofTs(tSeries, aryExp=aryExp)
      print *, 'getSN2ofTs() is ', f, ' which should be ', 16/6.0
      ! print *, 'getSN2ofTs() is ', getSN2ofTs(tSeries, aryExp=aryExp), ' which should be ', 16/6.0
      call error("err"//trim(errHead), '')
    end if

  end subroutine test_getSN2ofTs


  !---------------------------------------------------------------------
  subroutine test_findOptimumY()
  !---------------------------------------------------------------------
  ! Tests of findOptimumY()
  ! NOTE: 

    implicit none

    character(64)  :: errHead
    integer, parameter :: defArraySize = 7, DefPrecision = 5

    integer,      dimension(defArraySize) :: tSeriesInt
    real(single), dimension(defArraySize) :: tSeries, tSeriesRate, aryFracexp
    logical,      dimension(defArraySize) :: mask
    real(single) :: optimumY, lowerNbinRatio, expected
    integer(int32) :: optimumSize, expectedSize
    integer :: status

    !-------------------------------------------------------------------
    ! Start:

    !----------------------------------------------------------
    ! findOptimumY()
    !----------------------------------------------------------
    errHead = "findOptimumY001a"
    print *,"=== "//trim(errHead)//": Simple ()"

    ! Prepare a set of example test data
    ! tSeriesInt(:) = (/ 0, 1, 0, 2, 3, 4, -1 /)	! => *** Arithmetic exception:  - aborting
    tSeriesInt(:) = (/ 1, 2, 3, 4, 5, 8, 9 /)
    tSeries(:) = tSeriesInt
    ! Obviously the last 2 indices (values of 8 & 9) are the high-BGD region.
    ! Therefore the expected optimumY is tSeries(5)==5.
    expected = 5	! In reality, tiny bit larger than this.

    call findOptimumY(tSeries, optimumY, status=status)
      ! DEBUG: nbin= 7 curSN2=   1.5312500 in Start.
      ! DEBUG: nbin= 6 curSN2=   1.5652174 Removed-(i,Y)= 7   9.0000000
      ! DEBUG: nbin= 5 curSN2=   1.6666666 Removed-(i,Y)= 6   8.0000000
      ! DEBUG: nbin= 4 curSN2=   1.6000000 Removed-(i,Y)= 5   5.0000000

!print *, optimumY
    if ((.not. isNearlyEqual(optimumY, expected, DefPrecision)) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', optimumY, ' which should be ', expected
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "findOptimumY002a"
    print *,"=== "//trim(errHead)//": Simple (with optimumSize)"

    ! Prepare a set of example test data
    tSeriesInt(:) = (/ 1, 2, 3, 4, 5, 8, 9 /)
    tSeries(:) = tSeriesInt
    ! Obviously the last 2 indices (values of 8 & 9) are the high-BGD region.
    ! Therefore the expected optimumY is tSeries(5)==5.
    expected = 5	! In reality, tiny bit larger than this.
    expectedSize = 5

    call findOptimumY(tSeries, optimumY, optimumSize, status=status)
      ! DEBUG: nbin= 7 curSN2=   1.5312500 in Start.
      ! DEBUG: nbin= 6 curSN2=   1.5652174 Removed-(i,Y)= 7   9.0000000
      ! DEBUG: nbin= 5 curSN2=   1.6666666 Removed-(i,Y)= 6   8.0000000
      ! DEBUG: nbin= 4 curSN2=   1.6000000 Removed-(i,Y)= 5   5.0000000

    if ((.not. isNearlyEqual(optimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', optimumY, ' which should be ', expected
      print *, 'OptimumSize is ', optimumSize, ' which should be ',expectedSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "findOptimumY003a"
    print *,"=== "//trim(errHead)//": Simple (with some identical values)"

    ! Prepare a set of example test data
    tSeriesInt(:) = (/ 1, 2, 9, 9, 9, 9, 100 /)
    tSeries(:) = tSeriesInt
    ! Obviously the last 5 indices (values of 9 and 100) are the high-BGD region.
    ! Therefore the expected optimumY is tSeries(2)==2.
    expected = 2	! In reality, tiny bit larger than this.
    expectedSize = 2

    call findOptimumY(tSeries, optimumY, optimumSize, status=status)
      ! DEBUG: nbin= 7 curSN2=   0.3525180 in Start.
      ! DEBUG: nbin= 6 curSN2=   0.9230769 Removed-(i,Y)= 7   1.0000000E+02
      ! DEBUG: nbin= 5                     Removed-(i,Y)= 4   9.0000000
      ! DEBUG: nbin= 4                     Removed-(i,Y)= 5   9.0000000
      ! DEBUG: nbin= 3                     Removed-(i,Y)= 6   9.0000000
      ! DEBUG: nbin= 2 curSN2=   1.3333334 Removed-(i,Y)= 6   9.0000000
      ! DEBUG: nbin= 1 curSN2=   1.0000000 Removed-(i,Y)= 2   2.0000000

    if ((.not. isNearlyEqual(OptimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', OptimumY, ' which should be ', expected
      print *, 'OptimumSize is ', OptimumSize, ' which should be ',expectedSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "findOptimumY004a"
    print *,"=== "//trim(errHead)//": Simple (with some identical values, from the start)"

    ! Prepare a set of example test data
    tSeriesInt(:) = (/ 1, 2, 9, 9, 9, 9, 9 /)
    tSeries(:) = tSeriesInt
    ! Obviously the last 5 indices (values of 9) are the high-BGD region.
    ! Therefore the expected optimumY is tSeries(2)==2.
    expected = 2	! In reality, tiny bit larger than this.
    expectedSize = 2

    call findOptimumY(tSeries, optimumY, optimumSize, status=status)	! , doDump=.true.)
      !  DEBUG: nbin= 7 curSN2=   1.0208334 in Start.
      !  DEBUG: nbin= 6                     Removed-(i,Y)= 4   9.0000000
      !  DEBUG: nbin= 5                     Removed-(i,Y)= 5   9.0000000
      !  DEBUG: nbin= 4                     Removed-(i,Y)= 6   9.0000000
      !  DEBUG: nbin= 3                     Removed-(i,Y)= 7   9.0000000
      !  DEBUG: nbin= 2 curSN2=   1.3333334 Removed-(i,Y)= 7   9.0000000
      !  DEBUG: nbin= 1 curSN2=   1.0000000 Removed-(i,Y)= 2   2.0000000
      !
      ! NOTE if the identical values WERE NOT considered, the results would be:
      !  DEBUG: nbin= 7 curSN2=   1.0208334 in Start.
      !  DEBUG: nbin= 6 curSN2=   0.9230769 Removed-i= 3 (Y=   9.0000000 )
      !  DEBUG: nbin= 5 curSN2=   0.8333333 Removed-i= 4 (Y=   9.0000000 )
      !  DEBUG: nbin= 4 curSN2=   0.7619048 Removed-i= 5 (Y=   9.0000000 )
      !  DEBUG: nbin= 3 curSN2=   0.7500000 Removed-i= 6 (Y=   9.0000000 )
      !  DEBUG: nbin= 2 curSN2=   1.3333334 Removed-i= 7 (Y=   9.0000000 )
      !  DEBUG: nbin= 1 curSN2=   1.0000000 Removed-i= 2 (Y=   2.0000000 )
      ! therefore it would stop when nbin==6, which is wrong.

    if ((.not. isNearlyEqual(OptimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', OptimumY, ' which should be ', expected
      print *, 'OptimumSize is ', OptimumSize, ' which should be ',expectedSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "findOptimumY005a"
    print *,"=== "//trim(errHead)//": Simple (uniform)"

    ! Prepare a set of example test data
    tSeriesInt(:) = 2	! Uniform time-series.
    tSeries(:) = tSeriesInt
    expected = 2	! In reality, tiny bit larger than this.
    expectedSize = 7

    call findOptimumY(tSeries, optimumY, optimumSize, status=status)
      ! DEBUG: nbin= 7 curSN2=   3.5000000 in Start.
      ! DEBUG: nbin= 6                     Removed-(i,Y)= 2   2.0000000
      ! DEBUG: nbin= 5                     Removed-(i,Y)= 3   2.0000000
      ! DEBUG: nbin= 4                     Removed-(i,Y)= 4   2.0000000
      ! DEBUG: nbin= 3                     Removed-(i,Y)= 5   2.0000000
      ! DEBUG: nbin= 2                     Removed-(i,Y)= 6   2.0000000
      ! DEBUG: nbin= 1                     Removed-(i,Y)= 7   2.0000000
      !** : warning (regardLowestBinsAsOptimum), Failed to find ...

    if ((.not. isNearlyEqual(optimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', optimumY, ' which should be ', expected
      print *, 'OptimumSize is ', optimumSize, ' which should be ',expectedSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if
    ! if ( &	! It should fail.
    !     (status .ne. -2)) then
    !   print *, 'status is ', status, ' which should be 0.'
    !   call error("err"//trim(errHead), '')
    ! end if

    !----------------------------------------------------------
    errHead = "findOptimumY006a"
    print *,"=== "//trim(errHead)//": Simple (uniform except one high value)"

    ! Prepare a set of example test data
    tSeriesInt(:) = 2	! Uniform time-series.
    tSeriesInt(7) = 10	! with one exception.
    tSeries(:) = tSeriesInt
    expected = 2	! In reality, tiny bit larger than this.
    expectedSize = 6

    call findOptimumY(tSeries, optimumY, optimumSize, status=status)
      ! DEBUG: nbin= 7 curSN2=   2.2272727 in Start.
      ! DEBUG: nbin= 6 curSN2=   3.0000000 Removed-(i,Y)= 7  10.0000000
      ! DEBUG: nbin= 5                     Removed-(i,Y)= 2   2.0000000
      ! DEBUG: nbin= 4                     Removed-(i,Y)= 3   2.0000000
      ! DEBUG: nbin= 3                     Removed-(i,Y)= 4   2.0000000
      ! DEBUG: nbin= 2                     Removed-(i,Y)= 5   2.0000000
      ! DEBUG: nbin= 1                     Removed-(i,Y)= 6   2.0000000
      !** : warning (regardLowestBinsAsOptimum), Failed to find ...

    if ((.not. isNearlyEqual(optimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', optimumY, ' which should be ', expected
      print *, 'OptimumSize is ', optimumSize, ' which should be ',expectedSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if
    ! if ( &	! It should fail.
    !     (status .ne. -2)) then
    !   print *, 'status is ', status, ' which should be 0.'
    !   call error("err"//trim(errHead), '')
    ! end if

    !----------------------------------------------------------
    errHead = "findOptimumY007a"
    print *,"=== "//trim(errHead)//": Simple (random)"

    ! Prepare a set of example test data
    tSeriesInt(:) = (/ 5, 8, 1, 2, 9, 4, 3 /)
    tSeries(:) = tSeriesInt
    ! Obviously the 2 indices (values of 8 & 9) are the high-BGD region.
    ! Therefore the expected optimumY is tSeries(5)==5.
    expected = 5	! In reality, tiny bit larger than this.
    expectedSize = 5

    call findOptimumY(tSeries, optimumY, optimumSize, status=status)
      ! DEBUG: nbin= 7 curSN2=   1.5312500 in Start.
      ! DEBUG: nbin= 6 curSN2=   1.5652174 Removed-(i,Y)= 5   9.0000000
      ! DEBUG: nbin= 5 curSN2=   1.6666666 Removed-(i,Y)= 2   8.0000000
      ! DEBUG: nbin= 4 curSN2=   1.6000000 Removed-(i,Y)= 1   5.0000000

    if ((.not. isNearlyEqual(OptimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', OptimumY, ' which should be ', expected
      print *, 'OptimumSize is ', OptimumSize, ' which should be ',expectedSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if


    !----------------------------------------------------------
    errHead = "findOptimumY011a"
    print *,"=== "//trim(errHead)//": Simple (with mask)"

    ! Prepare a set of example test data
    tSeriesInt(:) = (/ 1, 2, 3, 4, 5, 8, 9 /)
    tSeries(:) = tSeriesInt
    ! Obviously the last 2 indices (values of 8 & 9) are the high-BGD region.
    mask(:) = .true.
    mask(5) = .false.
    ! The expected optimumY is tSeries(4)==4, after the mask is applied.
    expected = 4	! In reality, tiny bit larger than this.
    expectedSize = 4

    call findOptimumY(tSeries, optimumY, optimumSize, mask=mask, status=status)
      ! DEBUG: nbin= 6 curSN2=   1.3333334 in Start.
      ! DEBUG: nbin= 5 curSN2=   1.3888888 Removed-(i,Y)= 7   9.0000000
      ! DEBUG: nbin= 4 curSN2=   1.6000000 Removed-(i,Y)= 6   8.0000000
      ! DEBUG: nbin= 3 curSN2=   1.5000000 Removed-(i,Y)= 4   4.0000000

    if ((.not. isNearlyEqual(OptimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', OptimumY, ' which should be ', expected
      print *, 'OptimumSize is ', OptimumSize, ' which should be ',expectedSize
      print *, 'status is ', status, 'which should be 0.'
      call error("err"//trim(errHead), '')
    end if


    !----------------------------------------------------------
    errHead = "findOptimumY012a"
    print *,"=== "//trim(errHead)//": Simple (with mask of all False)"

    ! Prepare a set of example test data
    tSeriesInt(:) = (/ 1, 2, 3, 4, 5, 8, 9 /)
    tSeries(:) = tSeriesInt
    mask(:) = .false.
    ! The expected optimumY is tSeries(4)==4, after the mask is applied.
    expected = -4	! The returned status must be -4.
    expectedSize = 4

    call findOptimumY(tSeries, optimumY, optimumSize, mask=mask, status=status)	! warning (allMaskFalse) is raised.	! fracexp does nothing here as %style is 'none'


    if (status .ne. -4) then
      print *, 'status is ', status, 'which should be -4.'
      call error("err"//trim(errHead), '')
    end if


    !----------------------------------------------------------
    errHead = "findOptimumY021a"
    print *,"=== "//trim(errHead)//": Simple (with lowerNbinRatio; works)"

    ! Prepare a set of example test data
    tSeriesInt(:) = (/ 1, 2, 3, 4, 5, 8, 9 /)
    tSeries(:) = tSeriesInt
    ! Obviously the last 2 indices (values of 8 & 9) are the high-BGD region.
    lowerNbinRatio = 0.05	! Should not affect anything
    expected = 5.0	! In reality, tiny bit larger than this.
    expectedSize = 5

    call findOptimumY(tSeries, optimumY, optimumSize, lowerNbinRatio=lowerNbinRatio, status=status)
      ! DEBUG: nbin= 7 curSN2=   1.5312500 in Start.
      ! DEBUG: nbin= 6 curSN2=   1.5652174 Removed-(i,Y)= 7   9.0000000
      ! DEBUG: nbin= 5 curSN2=   1.6666666 Removed-(i,Y)= 6   8.0000000
      ! DEBUG: nbin= 4 curSN2=   1.6000000 Removed-(i,Y)= 5   5.0000000

    if ((.not. isNearlyEqual(OptimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', OptimumY, ' which should be ', expected
      print *, 'OptimumSize is ', OptimumSize, ' which should be ',expectedSize
      print *, 'status is ', status, 'which should be 0.'
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "findOptimumY022a"
    print *,"=== "//trim(errHead)//": Simple (with lowerNbinRatio; fails)"

    ! Prepare a set of example test data
    tSeriesInt(:) = (/ 1, 2, 3, 4, 5, 8, 9 /)
    tSeries(:) = tSeriesInt
    ! Obviously the last 2 indices (values of 8 & 9) are the high-BGD region.
    lowerNbinRatio = 0.99
    expected = -1.0
    expectedSize = -1.0

    call findOptimumY(tSeries, optimumY, optimumSize, lowerNbinRatio=lowerNbinRatio, status=status)
      ! DEBUG: nbin= 7 curSN2=   1.5312500 in Start.
      ! DEBUG: nbin= 6 curSN2=   1.5652174 Removed-(i,Y)= 7   9.0000000
      !:- Failed to find the optimum threshold.

    if ((OptimumY > 0) .or. &
        (optimumSize > 0) .or. &
        (status > -2)) then
      print *, 'OptimumY is ', OptimumY, ' which should be ', expected
      print *, 'OptimumSize is ', OptimumSize, ' which should be ',expectedSize
      print *, 'status is ', status, 'which should be 0.'
      call error("err"//trim(errHead), '')
    end if


    !----------------------------------------------------------
    errHead = "findOptimumY031a"
    print *,"=== "//trim(errHead)//": With tsRate"

    ! Prepare a set of example test data
    tSeriesInt(:) = (/ 1, 2, 3, 4, 5, 80, 90 /)
    tSeries(:) = tSeriesInt
    tSeries(6) = tSeries(6)*0.12
    tSeries(7) = tSeries(7)*0.08
    tSeriesRate(:) = tSeriesInt
    ! Obviously the last 2 indices (values of 8 & 9) are the high-BGD region.
    ! Therefore the expected optimumY is tSeries(5)==5.
    expected = 5	! In reality, tiny bit larger than this.
    expectedSize = 5

    call findOptimumY(tSeries, optimumY, optimumSize, tsRate=tSeriesRate, status=status)
      ! DEBUG: nbin= 7 curSN2=   0.8503145 in Start.
      ! DEBUG: nbin= 6 curSN2=   1.0656260 Removed-(i,Y)= 7  90.0000000
      ! DEBUG: nbin= 5 curSN2=   1.6666666 Removed-(i,Y)= 6  80.0000000
      ! DEBUG: nbin= 4 curSN2=   1.6000000 Removed-(i,Y)= 5   5.0000000
      ! DEBUG: nbin= 3 curSN2=   1.5000000 Removed-(i,Y)= 4   4.0000000
      ! DEBUG: nbin= 2 curSN2=   1.3333334 Removed-(i,Y)= 3   3.0000000
      ! DEBUG: nbin= 1 curSN2=   1.0000000 Removed-(i,Y)= 2   2.0000000

    if ((.not. isNearlyEqual(optimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', optimumY, ' which should be ', expected
      print *, 'OptimumSize is ', optimumSize, ' which should be ',expectedSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "findOptimumY032a"
    print *,"=== "//trim(errHead)//": With tsRate (2)"

    ! Prepare a set of example test data
    tSeriesInt(:) = (/ 1, 2, 3, 4, 5, 80, 90 /)
    tSeries(:) = tSeriesInt
    tSeries(6) = tSeries(6)*0.05
    tSeries(7) = tSeries(7)*0.03
    tSeriesRate(:) = tSeriesInt
    ! Obviously the last 2 indices (values of 8 & 9) are the high-BGD region.
    ! Therefore the expected optimumY is tSeries(5)==5.
    expected = 5	! In reality, tiny bit larger than this.
    expectedSize = 5

    call findOptimumY(tSeries, optimumY, optimumSize, tsRate=tSeriesRate, status=status)
      ! DEBUG: nbin= 7 curSN2=   1.1892350 in Start.
      ! DEBUG: nbin= 6 curSN2=   1.3422369 Removed-(i,Y)= 7  90.0000000
      ! DEBUG: nbin= 5 curSN2=   1.6666666 Removed-(i,Y)= 6  80.0000000
      ! DEBUG: nbin= 4 curSN2=   1.6000000 Removed-(i,Y)= 5   5.0000000

    if ((.not. isNearlyEqual(optimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', optimumY, ' which should be ', expected
      print *, 'OptimumSize is ', optimumSize, ' which should be ',expectedSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "findOptimumY033a"
    print *,"=== "//trim(errHead)//": With aryFracexp"

    ! Prepare a set of example test data
    aryFracexp = 1.0
    aryFracexp(6) = 0.05
    aryFracexp(7) = 0.03
    tSeriesInt(:) = (/ 1, 2, 3, 4, 5, 80, 90 /)
    tSeries(:) = tSeriesInt
    tSeries(6) = tSeries(6)*aryFracexp(6)
    tSeries(7) = tSeries(7)*aryFracexp(7)
    tSeriesRate(:) = tSeriesInt
    ! Obviously the last 2 indices (values of 8 & 9) are the high-BGD region.
    ! Therefore the expected optimumY is tSeries(5)==5.
    expected = 5	! In reality, tiny bit larger than this.
    expectedSize = 5

    call findOptimumY(tSeries, optimumY, optimumSize=optimumSize, aryFracexp=aryFracexp, status=status)

    if ((.not. isNearlyEqual(optimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', optimumY, ' which should be ', expected
      print *, 'OptimumSize is ', optimumSize, ' which should be ',expectedSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "findOptimumY034a"
    print *,"=== "//trim(errHead)//": With tsRate and aryFracexp"

    ! Prepare a set of example test data
    aryFracexp = 1.0
    aryFracexp(6) = 0.05
    aryFracexp(7) = 0.03
    tSeriesInt(:) = (/ 1, 2, 3, 4, 5, 80, 90 /)
    tSeries(:) = tSeriesInt
    tSeries(6) = tSeries(6)*aryFracexp(6)
    tSeries(7) = tSeries(7)*aryFracexp(7)
    ! Obviously the last 2 indices (values of 8 & 9) are the high-BGD region.
    ! Therefore the expected optimumY is tSeries(5)==5.
    expected = 5	! In reality, tiny bit larger than this.
    expectedSize = 5

    call findOptimumY(tSeries, optimumY, optimumSize=optimumSize &
       , tsRate=tSeriesRate, aryFracexp=aryFracexp, status=status)

    if ((.not. isNearlyEqual(optimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', optimumY, ' which should be ', expected
      print *, 'OptimumSize is ', optimumSize, ' which should be ',expectedSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if

    !!! What is the expected return is not obvious...
    !
    ! !----------------------------------------------------------
    ! errHead = "findOptimumY035a"
    ! print *,"=== "//trim(errHead)//": With aryFracexp, flat tSeries"
    ! 
    ! ! Prepare a set of example test data
    ! tSeries(:) = 1.0
    ! tSeriesInt(:) = (/ 3, 1, 2, 8, 9, 5, 4 /)
    ! aryFracexp(:) = 1.0/sqrt(real(tSeriesInt))	! Optimum => 9.5
    ! aryFracexp(:) = 1.0/tSeriesInt**2		! No Optimum (Status == -2)
    ! aryFracexp(:) = 1.0/tSeriesInt		! Optimum => 2.5
    ! ! Obviously i=4,5 (values of 8 & 9) are the high-BGD region.
    ! ! Therefore the expected optimumY is tSeries(5)==5.
    ! expected = 5	! In reality, tiny bit larger than this.
    ! expectedSize = 5
    ! 
    ! call findOptimumY(tSeries, optimumY, optimumSize=optimumSize &
    !    , aryFracexp=aryFracexp, status=status)
    ! 
    ! if ((.not. isNearlyEqual(optimumY, expected, DefPrecision)) .or. &
    !     (optimumSize .ne. expectedSize) .or. &
    !     (status .ne. 0)) then
    !   print *, 'OptimumY is ', optimumY, ' which should be ', expected
    !   print *, 'OptimumSize is ', optimumSize, ' which should be ',expectedSize
    !   print *, 'status is ', status, ' which should be 0.'
    !   call error("err"//trim(errHead), '')
    ! end if


  end subroutine test_findOptimumY


  !---------------------------------------------------------------------
  subroutine test_calcThresholdFromAny()
  !---------------------------------------------------------------------
  ! Tests of calcThresholdFromAny()
  ! NOTE: 

    implicit none

    character(64)  :: errHead
    integer, parameter :: defArraySize = 7, DefPrecision = 5

    integer,      dimension(defArraySize) :: tSeriesInt
    real(single), dimension(defArraySize) :: aryX, aryY, aryFracexp
    real(single) :: optimumY, lowerNbinRatio, expected
    integer(int32) :: optimumSize, maskSize, expectedSize, expectedMaskSize
    integer :: status, i
    type(InfoFracexpT)      :: fracexp
    type(ArgTimesStartEndT) :: timeStartEnd

    !-------------------------------------------------------------------
    ! Start:

    aryX(:) =     (/ -2, 0, 1, 2, 3, 4, 5 /)

    timeStartEnd%withStart = .true.
    timeStartEnd%withEnd   = .true.
    timeStartEnd%start = 0.0
    timeStartEnd%end   = 3.5

    fracexp%style = 'none'
    fracexp%doUse = .true.
    fracexp%colName = 'FRACEXP'
    fracexp%lowerThreshold = 0.4
    allocate(fracexp%ary(defArraySize))

    !----------------------------------------------------------
    ! calcThresholdFromAny()
    !----------------------------------------------------------
    errHead = "calcThresholdFromAny001a"
    print *,"=== "//trim(errHead)//": Simple ()"

    ! Prepare a set of example test data
    tSeriesInt(:) = (/ 3, 1, 2, 8, 9, 5, 4 /)
    aryY(:) = tSeriesInt	! Counts (or Rate)
    ! Obviously the values of 8 & 9 are the high-BGD region.
    ! Therefore the expected optimumY is tSeries(6)==5.
    expected = 5.0

    call calcThresholdFromAny(aryX, aryY, optimumY, status=status)
      ! DEBUG: nbin= 7 curSN2=   1.5312500 in Start.
      ! DEBUG: nbin= 6 curSN2=   1.5652174 Removed-(i,Y)= 5   9.0000000
      ! DEBUG: nbin= 5 curSN2=   1.6666666 Removed-(i,Y)= 4   8.0000000
      ! DEBUG: nbin= 4 curSN2=   1.6000000 Removed-(i,Y)= 6   5.0000000

    if ((.not. isNearlyEqual(OptimumY, expected, DefPrecision)) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', OptimumY, ' which should be ', expected
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "calcThresholdFromAny002a"
    print *,"=== "//trim(errHead)//": Simple (with optimumSize)"

    ! Prepare a set of example test data
    tSeriesInt(:) = (/ 3, 1, 2, 8, 9, 5, 4 /)
    aryY(:) = tSeriesInt	! Counts (or Rate)
    ! Obviously the values of 8 & 9 are the high-BGD region.
    ! Therefore the expected optimumY is tSeries(6)==5.
    expected = 5.0
    expectedSize = 5
    expectedMaskSize = 7

    call calcThresholdFromAny(aryX, aryY, optimumY, optimumSize, maskSize &
       , fracexp=fracexp, status=status)	! fracexp does nothing here as %style is 'none'
      ! DEBUG: nbin= 7 curSN2=   1.5312500 in Start.
      ! DEBUG: nbin= 6 curSN2=   1.5652174 Removed-(i,Y)= 5   9.0000000
      ! DEBUG: nbin= 5 curSN2=   1.6666666 Removed-(i,Y)= 4   8.0000000
      ! DEBUG: nbin= 4 curSN2=   1.6000000 Removed-(i,Y)= 6   5.0000000

    if ((.not. isNearlyEqual(OptimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (maskSize .ne. expectedMaskSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', optimumY, ' which should be ', expected
      print *, 'OptimumSize is ', optimumSize, ' which should be ',expectedSize
      print *, 'OptimumMaskSize is ', maskSize, ' which should be ',expectedMaskSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if


    !----------------------------------------------------------
    errHead = "calcThresholdFromAny011a"
    print *,"=== "//trim(errHead)//": With timeStartEnd"

    ! Prepare a set of example test data
    !aryX(:) ==    (/ -2, 0, 1, 2, 3, 4, 5 /)
    tSeriesInt(:) = (/ 3, 1, 2, 8, 9, 5, 4 /)
    ! MASK   ==     (/ F, T, T, T, T, F, F /)
    aryY(:) = tSeriesInt	! Counts (or Rate)
    timeStartEnd%start = 0.0
    timeStartEnd%end   = 3.5
    expected = 2.0
    expectedSize = 2
    expectedMaskSize = 4

    call calcThresholdFromAny(aryX, aryY, optimumY, optimumSize, maskSize &
       , fracexp=fracexp, timeStartEnd=timeStartEnd, status=status)	! fracexp does nothing here as %style is 'none'
      ! DEBUG: nbin= 4 curSN2=   0.8000000 in Start.
      ! DEBUG: nbin= 3 curSN2=   0.8181818 Removed-(i,Y)= 5   9.0000000
      ! DEBUG: nbin= 2 curSN2=   1.3333334 Removed-(i,Y)= 4   8.0000000
      ! DEBUG: nbin= 1 curSN2=   1.0000000 Removed-(i,Y)= 3   2.0000000

    if ((.not. isNearlyEqual(OptimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (maskSize .ne. expectedMaskSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', OptimumY, ' which should be ', expected
      print *, 'OptimumSize is ', optimumSize, ' which should be ',expectedSize
      print *, 'OptimumMaskSize is ', maskSize, ' which should be ',expectedMaskSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "calcThresholdFromAny012a"
    print *,"=== "//trim(errHead)//": With timeStartEnd (out of range, ie, all False)"

    ! Prepare a set of example test data
    tSeriesInt(:) = (/ 3, 1, 2, 8, 9, 5, 4 /)
    ! MASK   ==     (/ F, F, F, F, F, F, F /)
    aryY(:) = tSeriesInt	! Counts (or Rate)
    timeStartEnd%start = 99.0
    timeStartEnd%end   = 99.5
    ! expected = 2.0
    ! expectedSize = 2
    ! expectedMaskSize = 4

    call calcThresholdFromAny(aryX, aryY, optimumY, optimumSize, maskSize &
       , fracexp=fracexp, timeStartEnd=timeStartEnd, status=status)	! warning (noSignificantBin) is raised.	! fracexp does nothing here as %style is 'none'

    if (status .ne. -4) then
      print *, 'status is ', status, ' which should be -4.'
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "calcThresholdFromAny021a"
    print *,"=== "//trim(errHead)//": Simple (flat fracexp, style=calc)"

    ! Prepare a set of example test data
    tSeriesInt(:) = (/ 3, 1, 2, 8, 9, 5, 4 /)
    aryY(:) = tSeriesInt	! Counts (or Rate)

    fracexp%style = 'calc'
    fracexp%ary(:) = 0.5	! This does not affect the calculation, apart from the scaling factor in the returned value.

    expected = 10.0
    expectedSize = 5
    expectedMaskSize = 7

    call calcThresholdFromAny(aryX, aryY, optimumY, optimumSize, maskSize &
       , fracexp=fracexp, status=status)
      ! DEBUG: nbin= 7 curSN2=   0.3828125 in Start.
      ! DEBUG: nbin= 6 curSN2=   0.3913043 Removed-(i,Y)= 5  18.0000000
      ! DEBUG: nbin= 5 curSN2=   0.4166667 Removed-(i,Y)= 4  16.0000000
      ! DEBUG: nbin= 4 curSN2=   0.4000000 Removed-(i,Y)= 6  10.0000000

      ! (old...)
      ! DEBUG: nbin= 7 curSN2=   0.7656250 in Start.
      ! DEBUG: nbin= 6 curSN2=   0.7826087 Removed-(i,Y)= 5  18.0000000
      ! DEBUG: nbin= 5 curSN2=   0.8333333 Removed-(i,Y)= 4  16.0000000
      ! DEBUG: nbin= 4 curSN2=   0.8000000 Removed-(i,Y)= 6  10.0000000

    if ((.not. isNearlyEqual(OptimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (maskSize .ne. expectedMaskSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', OptimumY, ' which should be ', expected
      print *, 'OptimumSize is ', OptimumSize, ' which should be ',expectedSize
      print *, 'OptimumMaskSize is ', maskSize, ' which should be ',expectedMaskSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "calcThresholdFromAny022a"
    print *,"=== "//trim(errHead)//": Consistency (fracexp, style=calc, flat aryY)"

    ! Here we just check the consistency between findOptimumY() and calcThresholdFromAny().
    ! The latter is a wrapper of the former, so they have to be consistent.

    ! errHead = "findOptimumY035a"
    ! Prepare a set of example test data
    aryY(:) = 1.0
    tSeriesInt(:) = (/ 3, 1, 2, 8, 9, 5, 4 /)
    aryFracexp(:) = 1.0/tSeriesInt		! Optimum turns to be 2.5

    call findOptimumY(aryY, optimumY, optimumSize=optimumSize &
       , aryFracexp=aryFracexp, status=status)
    ! optimumY, optimumSize are set now.
    
    aryY(:) = 1.0	! Counts (or Rate)	! Flat input
    do i=1, size(tSeriesInt)
      fracexp%ary(i) = 1.0/tSeriesInt(i)	! Practically this gives the samy aryY(:) as the previous test once correction is applied.
    end do
    fracexp%style = 'calc'

    expected = optimumY
    expectedSize = optimumSize
    expectedMaskSize = 7

    call calcThresholdFromAny(aryX, aryY, optimumY, optimumSize, maskSize &
       , fracexp=fracexp, status=status)	! fracexp%style=='calc'
      ! DEBUG: nbin= 7 curSN2=   0.9068000 in Start.
      ! DEBUG: nbin= 6 curSN2=   0.9666783 Removed-(i,Y)= 5   9.0000000
      ! DEBUG: nbin= 5 curSN2=   1.0427222 Removed-(i,Y)= 4   8.0000000
      ! DEBUG: nbin= 4 curSN2=   1.0850694 Removed-(i,Y)= 6   5.0000000
      ! DEBUG: nbin= 3 curSN2=   1.1203704 Removed-(i,Y)= 7   4.0000000
      ! DEBUG: nbin= 2 curSN2=   1.1250000 Removed-(i,Y)= 1   3.0000000
      ! DEBUG: nbin= 1 curSN2=   1.0000000 Removed-(i,Y)= 3   2.0000000

      ! (old...)
      ! DEBUG: nbin= 7 curSN2=   1.5312500 in Start.
      ! DEBUG: nbin= 6 curSN2=   1.5652174 Removed-(i,Y)= 5   9.0000000
      ! DEBUG: nbin= 5 curSN2=   1.6666666 Removed-(i,Y)= 4   8.0000000
      ! DEBUG: nbin= 4 curSN2=   1.6000000 Removed-(i,Y)= 6   5.0000000

    if ((.not. isNearlyEqual(OptimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (maskSize .ne. expectedMaskSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', OptimumY, ' which should be ', expected
      print *, 'OptimumSize is ', OptimumSize, ' which should be ',expectedSize
      print *, 'OptimumMaskSize is ', maskSize, ' which should be ',expectedMaskSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if


    !----------------------------------------------------------
    errHead = "calcThresholdFromAny031a"
    print *,"=== "//trim(errHead)//": With lowerCutoffY"

    ! Prepare a set of example test data
    !aryX(:) ==    (/ -2, 0, 1, 2, 3, 4, 5 /)
    tSeriesInt(:) = (/ 0, 1, 2, 6, 6, 33, 33 /)
    ! Therefore, if all the bins are used, the resultant value should fall
    ! somewhere between 2 and 6, whereas if the bins of 2 or below are ignored
    ! ("calcThresholdFromAny032a"), the resultant value should fall
    ! somewhere between 6 and 33.
    aryY(:) = tSeriesInt	! Counts (or Rate)

    expected = 2.0
    expectedSize = 2
    expectedMaskSize = 6

    call calcThresholdFromAny(aryX, aryY, optimumY, optimumSize, maskSize &
       ,                   status=status)
      ! DEBUG: nbin= 6 curSN2=   0.4444444 in Start.
      ! DEBUG: nbin= 5                     Removed-(i,Y)= 7  33.0000000
      ! DEBUG: nbin= 4 curSN2=   1.0666667 Removed-(i,Y)= 7  33.0000000
      ! DEBUG: nbin= 3                     Removed-(i,Y)= 5   6.0000000
      ! DEBUG: nbin= 2 curSN2=   1.3333334 Removed-(i,Y)= 5   6.0000000
      ! DEBUG: nbin= 1 curSN2=   1.0000000 Removed-(i,Y)= 3   2.0000000

    if ((.not. isNearlyEqual(OptimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (maskSize .ne. expectedMaskSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', OptimumY, ' which should be ', expected
      print *, 'OptimumSize is ', OptimumSize, ' which should be ',expectedSize
      print *, 'OptimumMaskSize is ', maskSize, ' which should be ',expectedMaskSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "calcThresholdFromAny032a"
    print *,"=== "//trim(errHead)//": With lowerCutoffY"
    ! Look at "calcThresholdFromAny031a", do not change the variables below apart from expected.*!
    ! tSeriesInt(:) = (/ 0, 1, 2, 6, 6, 33, 33 /)

    expected = 6.0
    expectedSize = 3
    expectedMaskSize = 5

    call calcThresholdFromAny(aryX, aryY, optimumY, optimumSize, maskSize &
       , lowerCutoffY=1.0, status=status)
      ! DEBUG: nbin= 5 curSN2=   0.3125000 in Start.
      ! DEBUG: nbin= 4                     Removed-(i,Y)= 7  33.0000000
      ! DEBUG: nbin= 3 curSN2=   0.6428571 Removed-(i,Y)= 7  33.0000000
      ! DEBUG: nbin= 2                     Removed-(i,Y)= 5   6.0000000
      ! DEBUG: nbin= 1 curSN2=   0.5000000 Removed-(i,Y)= 5   6.0000000

    if ((maskSize .ne. expectedMaskSize) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (maskSize .ne. expectedMaskSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', OptimumY, ' which should be ', expected
      print *, 'OptimumSize is ', OptimumSize, ' which should be ',expectedSize
      print *, 'OptimumMaskSize is ', maskSize, ' which should be ',expectedMaskSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if


    !----------------------------------------------------------
    errHead = "calcThresholdFromAny041a"
    print *,"=== "//trim(errHead)//": With lowerNbinRatio"

    ! Prepare a set of example test data
    !aryX(:) ==    (/ -2, 0, 1, 2, 3, 4, 5 /)
    tSeriesInt(:) = (/ 3, 1, 2, 8, 9, 5, 4 /)
    ! MASK   ==     (/ T, T, T, T, T, F, F /)	! When time filtering is applied.
    aryY(:) = tSeriesInt	! Counts (or Rate)
    timeStartEnd%start = -2.5
    timeStartEnd%end   = 3.5
    lowerNbinRatio = 0.74	! Only 2 bins will be examined (out of 5)

    ! expected = 10.0
    ! expectedSize = 5
    expectedMaskSize = 5

    call calcThresholdFromAny(aryX, aryY, optimumY, optimumSize, maskSize &
       , timeStartEnd=timeStartEnd, lowerNbinRatio=lowerNbinRatio, status=status)
      ! DEBUG: nbin= 5 curSN2=   1.0869565 in Start.
      ! DEBUG: nbin= 4 curSN2=   1.1428572 Removed-(i,Y)= 5   9.0000000
      ! DEBUG: nbin= 3 curSN2=   1.5000000 Removed-(i,Y)= 4   8.0000000
      !:- Failed to find the optimum threshold.

    if ((maskSize .ne. expectedMaskSize) .or. &
        (status > -2)) then
      print *, 'OptimumY is ', OptimumY, ' which is arbitrary'
      print *, 'OptimumSize is ', OptimumSize, ' which is arbitrary'
      print *, 'OptimumMaskSize is ', maskSize, ' which should be ',expectedMaskSize
      print *, 'status is ', status, ' which should be negative.'
      call error("err"//trim(errHead), '')
    end if


    !----------------------------------------------------------
    errHead = "calcThresholdFromAny042a"
    print *,"=== "//trim(errHead)//": With lowerNbinRatio for step-function"

    ! Prepare a set of example test data
    !aryX(:) ==    (/ -2, 0, 1, 2, 3, 4, 5 /)
    tSeriesInt(:) = 2
    ! MASK   ==     (/ T, T, T, T, T, F, F /)	! When time filtering is applied.
    aryY(:) = tSeriesInt	! Counts (or Rate)
    aryY(6) = 9.997
    aryY(7) = 9.999
    lowerNbinRatio = 0.3

    expected = 2
    expectedSize = 5
    expectedMaskSize = 7

    call calcThresholdFromAny(aryX, aryY, optimumY, optimumSize, maskSize &
       , lowerNbinRatio=lowerNbinRatio, status=status)
      ! DEBUG: nbin= 7 curSN2=   1.6335511 in Start.
      ! DEBUG: nbin= 6 curSN2=   1.8002701 Removed-(i,Y)= 7   9.9989996
      ! DEBUG: nbin= 5 curSN2=   2.5000000 Removed-(i,Y)= 6   9.9969997
      ! DEBUG: nbin= 4                     Removed-(i,Y)= 2   2.0000000
      ! DEBUG: nbin= 3                     Removed-(i,Y)= 3   2.0000000
      ! DEBUG: nbin= 2                     Removed-(i,Y)= 4   2.0000000
      ! DEBUG: nbin= 1                     Removed-(i,Y)= 5   2.0000000
      !** : warning (regardLowestBinsAsOptimum), Failed to find ...

    if ((maskSize .ne. expectedMaskSize) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (maskSize .ne. expectedMaskSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', OptimumY, ' which should be ', expected
      print *, 'OptimumSize is ', OptimumSize, ' which should be ',expectedSize
      print *, 'OptimumMaskSize is ', maskSize, ' which should be ',expectedMaskSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if

    !----------------------------------------------------------
    errHead = "calcThresholdFromAny042b"
    print *,"=== "//trim(errHead)//": With lowerNbinRatio for step-function"
    ! See the above one (042a), and do not change the variables here!

    lowerNbinRatio = 0.4	! Because this ratio is not low enough, not all the bins are used for the calculation, therefore the subroutine returns the error (status).

    expected = 2
    expectedSize = 5
    expectedMaskSize = 7

    call calcThresholdFromAny(aryX, aryY, optimumY, optimumSize, maskSize &
       , lowerNbinRatio=lowerNbinRatio, status=status)
      ! DEBUG: nbin= 7 curSN2=   1.6335511 in Start.
      ! DEBUG: nbin= 6 curSN2=   1.8002701 Removed-(i,Y)= 7   9.9989996
      ! DEBUG: nbin= 5 curSN2=   2.5000000 Removed-(i,Y)= 6   9.9969997
      ! DEBUG: nbin= 4                     Removed-(i,Y)= 2   2.0000000
      ! DEBUG: nbin= 3                     Removed-(i,Y)= 3   2.0000000
      ! DEBUG: nbin= 2                     Removed-(i,Y)= 4   2.0000000

    if (&!(maskSize .ne. expectedMaskSize) .or. &
        !(optimumSize .ne. expectedSize) .or. &
        !(maskSize .ne. expectedMaskSize) .or. &
        (status > -2)) then
      !print *, 'OptimumY is ', OptimumY, ' which should be ', expected
      !print *, 'OptimumSize is ', OptimumSize, ' which should be ',expectedSize
      !print *, 'OptimumMaskSize is ', maskSize, ' which should be ',expectedMaskSize
      print *, 'status is ', status, ' which should be -3.'
      call error("err"//trim(errHead), '')
    end if

  end subroutine test_calcThresholdFromAny


  !---------------------------------------------------------------------
  subroutine test_calcThresholdFromRate()
  !---------------------------------------------------------------------
  ! Tests of calcThresholdFromRate()
  ! NOTE: 

    implicit none

    character(64)  :: errHead
    integer, parameter :: defArraySize = 7, DefPrecision = 5

    integer,      dimension(defArraySize) :: tSeriesInt
    real(single), dimension(defArraySize) :: aryX, aryY, aryFracexp
    real(single) :: optimumY, expected	! , lowerNbinRatio
    integer(int32) :: optimumSize, maskSize, expectedSize, expectedMaskSize
    integer :: status, i
    type(InfoFracexpT)      :: fracexp
    type(ArgTimesStartEndT) :: timeStartEnd

    !-------------------------------------------------------------------
    ! Start:

    aryX(:) =     (/ -2, 0, 1, 2, 3, 4, 5 /)

    timeStartEnd%withStart = .true.
    timeStartEnd%withEnd   = .true.
    timeStartEnd%start = 0.0
    timeStartEnd%end   = 3.5

    fracexp%style = 'none'
    fracexp%doUse = .true.
    fracexp%colName = 'FRACEXP'
    fracexp%lowerThreshold = 0.4
    allocate(fracexp%ary(defArraySize))

    !----------------------------------------------------------
    errHead = "calcThresholdFromRate022a"
    print *,"=== "//trim(errHead)//": Consistency (fracexp, style=calc, flat aryY)"

    ! Here we just check the consistency between findOptimumY() and calcThresholdFromRate().
    ! The latter is a wrapper of the former, so they have to be consistent.

    ! errHead = "findOptimumY035a"
    ! Prepare a set of example test data
    aryY(:) = 1.0
    tSeriesInt(:) = (/ 3, 1, 2, 8, 9, 5, 4 /)
    aryFracexp(:) = 1.0/tSeriesInt		! Optimum turns to be 2.5

    call findOptimumY(aryY, optimumY, optimumSize=optimumSize &
       , aryFracexp=aryFracexp, status=status)
    ! optimumY, optimumSize are set now.
    
    aryY(:) = 1.0	! Counts (or Rate)	! Flat input
    do i=1, size(tSeriesInt)
      fracexp%ary(i) = 1.0/tSeriesInt(i)	! Practically this gives the samy aryY(:) as the previous test once correction is applied.
    end do
    fracexp%style = 'calc'

    expected = optimumY
    expectedSize = optimumSize
    expectedMaskSize = 7

    call calcThresholdFromRate(aryX, aryY, optimumY, optimumSize, maskSize &
       , fracexp=fracexp, status=status)	! fracexp%style=='calc'
      ! DEBUG: nbin= 7 curSN2=   0.9068000 in Start.
      ! DEBUG: nbin= 6 curSN2=   0.9666783 Removed-(i,Y)= 5   9.0000000
      ! DEBUG: nbin= 5 curSN2=   1.0427222 Removed-(i,Y)= 4   8.0000000
      ! DEBUG: nbin= 4 curSN2=   1.0850694 Removed-(i,Y)= 6   5.0000000
      ! DEBUG: nbin= 3 curSN2=   1.1203704 Removed-(i,Y)= 7   4.0000000
      ! DEBUG: nbin= 2 curSN2=   1.1250000 Removed-(i,Y)= 1   3.0000000
      ! DEBUG: nbin= 1 curSN2=   1.0000000 Removed-(i,Y)= 3   2.0000000

      ! (old...)
      ! DEBUG: nbin= 7 curSN2=   1.5312500 in Start.
      ! DEBUG: nbin= 6 curSN2=   1.5652174 Removed-(i,Y)= 5   9.0000000
      ! DEBUG: nbin= 5 curSN2=   1.6666666 Removed-(i,Y)= 4   8.0000000
      ! DEBUG: nbin= 4 curSN2=   1.6000000 Removed-(i,Y)= 6   5.0000000

    if ((.not. isNearlyEqual(OptimumY, expected, DefPrecision)) .or. &
        (optimumSize .ne. expectedSize) .or. &
        (maskSize .ne. expectedMaskSize) .or. &
        (status .ne. 0)) then
      print *, 'OptimumY is ', OptimumY, ' which should be ', expected
      print *, 'OptimumSize is ', OptimumSize, ' which should be ',expectedSize
      print *, 'OptimumMaskSize is ', maskSize, ' which should be ',expectedMaskSize
      print *, 'status is ', status, ' which should be 0.'
      call error("err"//trim(errHead), '')
    end if

  end subroutine test_calcThresholdFromRate

end program testbkgoptrateaux01

